/* -*- Mode: C; indent-tabs-mode: nil; c-basic-offset: 8-*- */
/*
 *This file is part of MlView
 *
 *MlView is free software; you can redistribute 
 *it and/or modify it under the terms of 
 *the GNU General Public License as published 
 *by the Free Software Foundation; either version 2, 
 *or (at your option) any later version.
 *
 *MlView is distributed in the hope that 
 *it will be useful, but WITHOUT ANY WARRANTY; 
 *without even the implied warranty of 
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *See the GNU General Public License for more details.
 *
 *You should have received a copy of the 
 *GNU General Public License along with MlView; 
 *see the file COPYING. If not, write to the Free Software Foundation, 
 *Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 *Copyright 2001-2002 Dodji Seketeli.
 */

#include <libxml/tree.h>
#include <libxml/valid.h>

#include "mlview-xml-document.h"
#include "mlview-parsing-utils.h"
#include "mlview-settings-management.h"
#include "mlview-file-selection.h"
#include "mlview-file-descriptor.h"
#include "mlview-global-settings.h"

/**
 *@file
 *The definition of the #MlViewXMLDocument class.
 *
 *This is a wrapper of the "tree" module
 *of daniel veillard's libxml.
 *That way, all the real actions made on 
 *the underlying xml data structure
 *are made through this class.
 *This class then emits signals to
 *notify it's state to all the views that are
 *in relation with it instances. Of course, this
 *class should not know anything about the views that holds it ...
 *
 */

struct _MlViewXMLDocumentPrivate
{
	MlViewFileDescriptor *file_desc ;
	/*MlViewXMLDocTreeView * tree_view ;*/

	/*the underlying xml doc data structure*/
	xmlDocPtr xml_doc ;

	/*the current xml application context*/
	MlViewAppContext * app_context ;
        
        guint ref_count ;
} ;


enum 
{
	DOCUMENT_CHANGED = 0 ,
	NODE_CUT,
	CHILD_NODE_ADDED,
	PREV_SIBLING_NODE_INSERTED,
	NEXT_SIBLING_NODE_INSERTED,
	CONTENT_CHANGED,
	NAME_CHANGED,
	NODE_CHANGED,
	FILE_PATH_CHANGED,
	NUMBER_OF_SIGNALS
} ;

#define PRIVATE(mlview_xml_doc) ((mlview_xml_doc)->private)
#define MLVIEW_XML_DOCUMENT_DEFAULT_VALIDATION_ON FALSE ;


static void 
mlview_xml_document_do_interactive_validation_if_needed 
(MlViewXMLDocument * a_doc) ;

static void 
mlview_xml_document_destroy (GtkObject *a_xml_doc) ;

#define MLVIEW_XML_DOCUMENT_NODE_CLIPBOARD_SIZE 128
static GtkObjectClass *gv_parent_class = NULL ;
static guint gv_mlview_xml_document_signals [NUMBER_OF_SIGNALS] = {0} ;
static xmlNode * gv_clipboard[MLVIEW_XML_DOCUMENT_NODE_CLIPBOARD_SIZE] = {0} ;
static guint gv_clipboard_index = 0 ;
static guint gv_clipboard_ref_count = 0 ;


static void
mlview_xml_document_destroy_clipboard (void) ;

static void
mlview_xml_document_ref_clipboard (void) ;

static gboolean
mlview_xml_document_unref_clipboard (void) ;

static void
mlview_xml_document_do_interactive_validation_if_needed 
(MlViewXMLDocument * a_doc) ;

static void
mlview_xml_document_class_init (MlViewXMLDocumentClass * a_klass) ;

static void
mlview_xml_document_init (MlViewXMLDocument * a_xml_doc) ;

/*======================================================
 *private helper functions
 *=====================================================*/


/**
 *Destroys the instances of xmlNode
 *that are present in the clipboard.
 *
 */
static void
mlview_xml_document_destroy_clipboard (void)
{
        guint i = 0 ;

        for (i = 0 ; 
             i < gv_clipboard_index ; 
             i++)
        {
                if (gv_clipboard[i])
                {
                        xmlFreeNode (gv_clipboard[i]) ;
                        gv_clipboard[i] = NULL ;
                }
        }

        gv_clipboard_index = 0 ;
}


/**
 *Increments the ref counter of the
 *clipboard.
 */
static void
mlview_xml_document_ref_clipboard (void)
{
        gv_clipboard_ref_count ++ ;
}

/**
 *Decrements the ref count of the clipboard.
 *If the ref count reaches zero, the clipboard
 *if destroyed.
 *
 *@return TRUE if the clipboard got destroyed, FALSE
 *otherwise.
 */
static gboolean
mlview_xml_document_unref_clipboard (void)
{
        if (gv_clipboard_ref_count > 0)
                gv_clipboard_ref_count -- ;

        if (gv_clipboard_ref_count == 0)
        {
                mlview_xml_document_destroy_clipboard () ;
                return TRUE ;
        }
        
        return FALSE ;
}


/**
 *Validates the document if the settings allows
 *it.
 *@param a_doc the "this" pointer.
 */
static void
mlview_xml_document_do_interactive_validation_if_needed 
(MlViewXMLDocument * a_doc)
{
	gchar * validation_is_on = NULL ;

	g_return_if_fail(a_doc != NULL) ;
	g_return_if_fail(PRIVATE(a_doc) != NULL) ;
	g_return_if_fail(PRIVATE(a_doc)->xml_doc != NULL) ;

	if (PRIVATE (a_doc)->app_context)
		validation_is_on = 
			mlview_app_context_get_settings_value 
			(PRIVATE (a_doc)->app_context,
			 MLVIEW_STG_K_IS_VALIDATION_ON) ;
	
	if (validation_is_on
	    && (strcmp (validation_is_on, MLVIEW_STG_V_YES) == 0)) 
	{
		mlview_xml_document_validate (a_doc) ;
	}
}


/***********************************************
 *private gtk framework methods						
 ***********************************************/

/**
 *Inits the MlViewXMLDocument class.
 *@param a_klass the class structure containing the vtable
 *and all the class wide data.
 */
static void
mlview_xml_document_class_init (MlViewXMLDocumentClass * a_klass)
{
	GtkObjectClass * object_class = NULL ;

	g_return_if_fail (a_klass != NULL) ;
	g_return_if_fail (MLVIEW_IS_XML_DOCUMENT_CLASS (a_klass)) ;

	object_class = GTK_OBJECT_CLASS (a_klass) ;
	gv_parent_class = gtk_type_class (GTK_TYPE_OBJECT) ;
	object_class->destroy = mlview_xml_document_destroy ;

	/*signal creation code*/
	gv_mlview_xml_document_signals[DOCUMENT_CHANGED] = 
		gtk_signal_new ("document-changed",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (MlViewXMLDocumentClass,
						   document_changed),
				gtk_marshal_NONE__NONE,
				GTK_TYPE_NONE, 0) ;

	gv_mlview_xml_document_signals[NODE_CUT] = 
		gtk_signal_new ("node-cut",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (MlViewXMLDocumentClass, 
						   node_cut),
				gtk_marshal_NONE__POINTER_POINTER,
				GTK_TYPE_NONE, 2, 
				GTK_TYPE_POINTER, GTK_TYPE_POINTER) ;

	gv_mlview_xml_document_signals [CHILD_NODE_ADDED] = 
		gtk_signal_new ("child-node-added",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (MlViewXMLDocumentClass, 
						   child_node_added),
				gtk_marshal_NONE__POINTER_POINTER,
				GTK_TYPE_NONE, 2, 
				GTK_TYPE_POINTER, GTK_TYPE_POINTER) ;

	gv_mlview_xml_document_signals [PREV_SIBLING_NODE_INSERTED] = 
		gtk_signal_new ("prev-sibling-node-inserted",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (MlViewXMLDocumentClass, 
						   prev_sibling_node_inserted),
				gtk_marshal_NONE__POINTER_POINTER,
				GTK_TYPE_NONE, 2, 
				GTK_TYPE_POINTER, GTK_TYPE_POINTER) ;

	gv_mlview_xml_document_signals [NEXT_SIBLING_NODE_INSERTED] = 
		gtk_signal_new ("next-sibling-node-inserted",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (MlViewXMLDocumentClass, 
						   next_sibling_node_inserted),
				gtk_marshal_NONE__POINTER_POINTER,
				GTK_TYPE_NONE, 2, 
				GTK_TYPE_POINTER, GTK_TYPE_POINTER) ;

	gv_mlview_xml_document_signals [CONTENT_CHANGED] = 
		gtk_signal_new ("content-changed",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (MlViewXMLDocumentClass, 
						   content_changed),
				gtk_marshal_NONE__POINTER,
				GTK_TYPE_NONE, 1, 
				GTK_TYPE_POINTER) ;

	gv_mlview_xml_document_signals [NAME_CHANGED] = 
		gtk_signal_new ("name-changed",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (MlViewXMLDocumentClass, 
						   name_changed),
				gtk_marshal_NONE__POINTER,
				GTK_TYPE_NONE, 1, 
				GTK_TYPE_POINTER) ;

	gv_mlview_xml_document_signals [NODE_CHANGED] = 
		gtk_signal_new ("node-changed",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (MlViewXMLDocumentClass, 
						   node_changed),
				gtk_marshal_NONE__POINTER,
				GTK_TYPE_NONE, 1, 
				GTK_TYPE_POINTER) ;

	gv_mlview_xml_document_signals [FILE_PATH_CHANGED] =
		gtk_signal_new ("file-path-changed",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (MlViewXMLDocumentClass, 
						   file_path_changed),
				gtk_marshal_NONE__NONE,
				GTK_TYPE_NONE, 0) ;

	gtk_object_class_add_signals (object_class, 
				      gv_mlview_xml_document_signals, 
				      NUMBER_OF_SIGNALS) ;

	a_klass->document_changed = NULL ;
	a_klass->node_cut = NULL ;
	a_klass->child_node_added = NULL ;
	a_klass->node_changed = NULL ;
	a_klass->content_changed = NULL ;
	a_klass->name_changed = NULL ;
}

/**
 *The init of the class instance.
 *@param a_xml_doc the newly built "this pointer".
 */
static void
mlview_xml_document_init (MlViewXMLDocument * a_xml_doc)

{
	g_return_if_fail (a_xml_doc != NULL) ;
	g_return_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_doc)) ;

	PRIVATE (a_xml_doc) = g_malloc0 (sizeof (MlViewXMLDocumentPrivate)) ;
}

/*=============================================
 *public methods
 *=============================================*/


/**
 *The actual #MlViewXMLDocument type ID builder.
 *@return the type ID of the #MlViewXMLDocument class.
 */
guint 
mlview_xml_document_get_type (void)
{
	static guint mlview_xml_document_type = 0 ;
	
	if (!mlview_xml_document_type) 
	{
		static const GtkTypeInfo t_info = 
		{
			"MlViewXMLDocument",
			sizeof (MlViewXMLDocument),
			sizeof (MlViewXMLDocumentClass),
			(GtkClassInitFunc) mlview_xml_document_class_init,
			(GtkObjectInitFunc) mlview_xml_document_init,
			NULL, NULL, NULL
		} ;
		mlview_xml_document_type = 
			gtk_type_unique (GTK_TYPE_OBJECT, &t_info) ;
	}

	return mlview_xml_document_type ;
}


/**
 *The class instance builder.
 *Builds a new instance of the #MlViewXMLDocument. Initialyzes
 *it with a native instance of xmlDoc * (see libxml2 to learn more
 *about xmlDoc*)
 *
 *@param a_xml_doc the native libxml2 xmlDocPtr.
 *@param a_context the mlview application context.
 *@return the newly created instance of MlViewXMLDocument.
 */
MlViewXMLDocument *
mlview_xml_document_new (xmlDocPtr a_xml_doc, MlViewAppContext *a_context)
{
	MlViewXMLDocument *result = NULL ;
	xmlDocPtr xml_doc=NULL ;
	xmlNodePtr xml_node ;

	result = gtk_type_new (MLVIEW_TYPE_XML_DOCUMENT) ;

	xml_doc = a_xml_doc ;

	if (xml_doc == NULL) 
	{
		xml_doc = xmlNewDoc ("1.0") ;
		xml_node = xmlNewNode (NULL, "") ;
		xmlDocSetRootElement (xml_doc, xml_node) ;
	}

	PRIVATE (result)->xml_doc = xml_doc ;
	PRIVATE (result)->app_context = a_context ;

        mlview_xml_document_ref_clipboard () ;

	return result ;
}


/**
 *Associates the current instance of 
 *#MlViewXMLDocument with a new instance
 *of #MlViewApplicationContext.
 *
 *@param a_doc the current xml document.
 *@param a_context the current application 
 *context to set to this component.
 */
void
mlview_xml_document_set_app_context (MlViewXMLDocument * a_doc, 
				     MlViewAppContext * a_context)
{
	g_return_if_fail (a_doc != NULL) ;
	g_return_if_fail (PRIVATE (a_doc) != NULL) ;
	g_return_if_fail (a_context != NULL) ;
	g_return_if_fail (MLVIEW_IS_APP_CONTEXT (a_context)) ;


	PRIVATE (a_doc)->app_context = a_context ;
}


/**
 *A Getter of the application context associated to
 *this instance of #MlViewAppContext.
 *
 *@param a_doc the "this pointer".
 *@return the application context.
 */
MlViewAppContext *
mlview_xml_document_get_app_context (MlViewXMLDocument * a_doc) 
{
	g_return_val_if_fail (a_doc, NULL) ;
	g_return_val_if_fail (PRIVATE (a_doc) != NULL, NULL) ;
	
	return PRIVATE (a_doc)->app_context ;
}


/************************************
 *clipboard methods.
 ***********************************/

/**
 *Stores the a copy of the xmlNode given in argument into the node clipboard.
 *The cilpboard is a circular buffer. It cannot be overflowed. 
 *
 *@param a_visual_node the pointer to 
 *xmlNode to store in the clipboard. If it is a 
 *NULL pointer, a NULL pointer is actually stored into clipboard.
 */
void
mlview_xml_document_copy_node_to_clipboard (xmlNode *a_xml_node, 
					    xmlDoc *a_doc)
{
	g_return_if_fail (a_xml_node != NULL) ;

	g_return_if_fail (sizeof (gv_clipboard) / sizeof (xmlNodePtr) > 0) ;

        /*if buffer full, go to beginning of buffer.*/
	if (gv_clipboard_index >= sizeof (gv_clipboard) / sizeof (xmlNodePtr)) 
	{
		gv_clipboard_index = 0 ;
	}

        /*avoid mem leaks*/
        if (gv_clipboard[gv_clipboard_index])
        {
                xmlFreeNode (gv_clipboard[gv_clipboard_index]) ;
                gv_clipboard[gv_clipboard_index] = NULL ;
        }

	gv_clipboard[gv_clipboard_index] = 
		xmlDocCopyNode (a_xml_node, a_doc , TRUE) ;

	gv_clipboard_index ++ ;
}


/**
 *Gets the last node stored in the node clipboard.
 *
 *@param a_xml_doc the "this pointer".
 *@return a copy of the 
 *xmlNode that has been previously stored. 
 */
xmlNode *
mlview_xml_document_get_node_from_clipboard (xmlDoc * a_xml_doc)
{
	xmlNodePtr xml_node=NULL ;

	g_return_val_if_fail (sizeof(gv_clipboard)/sizeof(xmlNodePtr) > 0, 
			      NULL) ;

	if (gv_clipboard_index >= sizeof(gv_clipboard)/sizeof(xmlNodePtr)) 
	{
		gv_clipboard_index = 
			sizeof(gv_clipboard)/sizeof(xmlNodePtr) - 1 ;
	}

	xml_node = gv_clipboard[gv_clipboard_index - 1] ;

	if (xml_node != NULL) 
	{
		/*make a recursive copy of the node*/
		xml_node = 
			xmlDocCopyNode (xml_node, a_xml_doc, 1) ;
	}

	return xml_node ;
}


/*************************************
 *Other miscellaneous methods
 *************************************/

/**
 *Opens and loads the xml file contains in the filename given in parameter. 
 *
 *@param a_file_name the total file path of the file that contains 
 *the xml document to open
 *@param a_context the mlview application context.
 *the current instance of MlViewXMLDocument.
 *@return the newly loaded xml document.
 */
MlViewXMLDocument *
mlview_xml_document_open (gchar * a_file_name, 
			  MlViewAppContext * a_context)
{
	MlViewXMLDocument *result = NULL;
	xmlDoc * xml_doc= NULL ;

	xml_doc = mlview_parsing_utils_load_xml_file (a_file_name, a_context) ;
	g_return_val_if_fail (xml_doc != NULL , NULL) ;

	result = mlview_xml_document_new (xml_doc, a_context) ;

	mlview_xml_document_do_interactive_validation_if_needed (result) ;
	mlview_xml_document_set_file_path (result, a_file_name) ;

	return result;
}


/**
 *Associates a Dtd to the current xml document.
 *If the document already contains
 *a Dtd, this functions displays a dialog box 
 *to ask the user if he wants to overide
 *that Dtd. Depending on the answer of the user, 
 *this function does the association
 *or not. 
 *
 *@param a_doc the current mlview xml document.
 *@return 0 if the association has been done, 
 *1 if the user cancelled the operations, 
 */
gint
mlview_xml_document_associate_dtd_interactive (MlViewXMLDocument * a_doc)
{
	MlViewExternalSubsetDefinition * external_subset_def = NULL ;
	gint result = 0 ;

	g_return_val_if_fail (a_doc != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_doc) != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_doc)->app_context != NULL, -1) ;

	external_subset_def = 
		mlview_parsing_utils_let_user_choose_a_dtd 
		(PRIVATE (a_doc)->app_context, 
		 _("Choose a Dtd to associate to the document")) ;

	if ( external_subset_def ) 
	{
		result = 
			mlview_xml_document_associate_dtd (a_doc,
							   external_subset_def);
		mlview_external_subset_definition_destroy 
			(external_subset_def) ;
		return result ;
		
	} 
	else 
	{
		return 1 ;
	}
}


/**
 *Associates a the dtd designated by 
 *a_ext_subs_def to the current xml document.
 *Note that if the document contains is already 
 *associated to a dtd, the function asks
 *the user if she wants to overide it. 
 *If the user agrees, the functions frees the former
 *dtd and makes the new association. If not, nothing is done. 
 *
 *@return 0 if the association has been done, 1 
 *if the user cancelled the operations, 
 */
gint
mlview_xml_document_associate_dtd 
(MlViewXMLDocument * a_doc, 
 MlViewExternalSubsetDefinition * a_ext_subs_def)
{
	xmlDtd * dtd = NULL ;
	gboolean user_cancels = FALSE ;

	g_return_val_if_fail (a_doc != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_doc) != NULL, -1) ;

	/*
	 *if an external subset already exists, 
	 *ask the user if she wants to overide it
	 */
	if ( PRIVATE(a_doc)->xml_doc && PRIVATE(a_doc)->xml_doc->extSubset ) 
	{
		GnomeDialog * dialog = NULL ;
		GtkWidget * label = NULL ;
		gint button = 0 ;

		dialog = GNOME_DIALOG
			(gnome_dialog_new(_("Document has already has a Dtd"),
					  _("YES"),_("CANCEL"), NULL)) ;
		label = gtk_label_new 
			(_("This Document already has an associated Dtd.\nDo you want really want to associate another Dtd to this document ?")) ;

		gtk_box_pack_start (GTK_BOX(dialog->vbox),
				    label, TRUE,FALSE,0) ;

		gtk_widget_show_all (dialog->vbox) ;
		gnome_dialog_close_hides (dialog, FALSE) ;
		gnome_dialog_set_close (dialog, TRUE) ;
		button = gnome_dialog_run (dialog) ;

		switch ( button ) 
		{
		case 0:/*YES button*/
			break ;
		default :/*CANCEL Button, dialog box closed etc..*/
			user_cancels = TRUE ;
			break ;
		}
	}/*end if*/
	
	if ( user_cancels == TRUE )
		return 1 ;/*user cancelled operations*/

	/*if the user did not cancel the operations, let's do the association*/
	
	/*let's parse the external subset. 
	 *If parsing is OK, we associate that dtd to
	 *the document.
	 */
	if ( a_ext_subs_def->system_id  
	     &&  strcmp (a_ext_subs_def->system_id, "")
	     &&  (dtd = mlview_parsing_utils_load_a_dtd 
		  (a_ext_subs_def, 
		   PRIVATE(a_doc)->app_context)) ) 
        {
		
		if (PRIVATE(a_doc)->xml_doc->extSubset) 
			xmlFreeDtd(PRIVATE(a_doc)->xml_doc->extSubset) ;

		PRIVATE(a_doc)->xml_doc->extSubset = dtd ;

		dtd->doc = PRIVATE(a_doc)->xml_doc ;

		PRIVATE(a_doc)->xml_doc->standalone = 0 ;

		return 0 ; /*dtd parsing OK, association OK*/
	}

	return 2 ;/*dtd parsing failed*/
}


/**
 *Validates the document agains the dtd that is associated to it. 
 *
 *@param a_doc the document to validate.
 *@return O upon successfull completion, or an error code.
 */
gint
mlview_xml_document_validate (MlViewXMLDocument * a_doc)
{
	/*this defaul value means "there is no dtd to parse"*/
	gint validity_status=-2 ;

	g_return_val_if_fail (a_doc != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_doc) != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_doc)->xml_doc, -1) ;

	if ( PRIVATE(a_doc)->xml_doc->extSubset
	     || PRIVATE(a_doc)->xml_doc->intSubset) 
	{

		validity_status =
			mlview_parsing_utils_validate_dtd 
			(PRIVATE(a_doc)->xml_doc,
			 PRIVATE(a_doc)->xml_doc->extSubset,
			 PRIVATE(a_doc)->app_context) ;
	} 
	else 
	{

		mlview_app_context_warning 
			(PRIVATE (a_doc)->app_context, 
			 _("No DTD was provided. Could not validate the document"))  ;

		return validity_status ;
	}

	/*If the document is valid, tell the user.*/
	if (! validity_status )

		mlview_app_context_message (PRIVATE (a_doc)->app_context, 
					    _("The Document is valid :-)")) ;
	else
		mlview_app_context_warning (PRIVATE (a_doc)->app_context, 
					    _("The Document is not valid !"))  ;

	return validity_status ;
}


/**
 *Saves the document.
 *
 *@param a_doc the current instance of MlViewXMLDocument
 *@param a_file_path the path where to save the document.
 *@param a_check_overwrt 
 *@return the number of bytes saved or -1 in case of error.
 */
gint
mlview_xml_document_save (MlViewXMLDocument * a_doc, 
                          gchar * a_file_path,
                          gboolean a_check_overwrt)
{
	MlViewFileDescriptor * file_desc = NULL ;
        gboolean save_file = TRUE, is_reg_file = TRUE ;
        
	g_return_val_if_fail(a_doc != NULL, -1) ;
	g_return_val_if_fail(PRIVATE(a_doc) != NULL, -1) ;
        
        if (a_check_overwrt == TRUE)
        {
                file_desc = mlview_file_descriptor_new (a_file_path) ;
                g_return_val_if_fail (file_desc, -1) ;

                if (!mlview_file_descriptor_is_regular_file (file_desc, 
                                                             &is_reg_file)
                    && is_reg_file == TRUE)
                {
                        GnomeDialog *ok_cancel_dialog = NULL ;
                        gchar * str = NULL ;

                        str = 
                                g_strconcat 
                                (_("The file "),
                                 a_file_path,
                                 _(" already exists.\n"),
                                 _("Do you want to overwrite it ?.\n"),
                                 NULL) ;

                        ok_cancel_dialog = 
                                GNOME_DIALOG 
                                (gnome_ok_cancel_dialog (str, NULL, NULL)) ;

                        if (str)
                        {
                                g_free (str) ;
                                str = NULL ;
                        }
                
                        if (ok_cancel_dialog)
                        {
                                gint button = 0 ;

                                gnome_dialog_close_hides (ok_cancel_dialog, 
                                                          FALSE) ;

                                button = gnome_dialog_run_and_close 
                                        (ok_cancel_dialog) ;

                                switch (button)
                                {
                                case 0: /*The OK Button was pressed*/
                                        save_file = TRUE ;
                                        break ;

                                case 1:/*The cancel Button was pressed*/
                                case -1:
                                default :
                                        save_file = FALSE ;
                                        break ;
                                }
                        }
                }
        }

        if (file_desc)
        {
                mlview_file_descriptor_destroy (file_desc) ;
                file_desc = NULL ;
        }

        if (save_file == TRUE)
        {
                gint res = 0 ;
                xmlDoc * xml_doc = NULL ;

                xml_doc = mlview_xml_document_get_xml_document (a_doc) ;
                g_return_val_if_fail (xml_doc, -1) ;

                res = mlview_parsing_utils_save_xml_doc 
                        (xml_doc, a_file_path, PRIVATE(a_doc)->app_context) ;

                if (res > 0)
                        mlview_xml_document_set_file_path (a_doc, a_file_path) ;
                
                return res ;
        }

        return 0 ;
}


/**
 *Clones the current instance of MlViewXMLDocument. 
 *@param a_original the current instance of MlViewXMLDocument.
 *@return the cloned instance of #MlViewXMLDocument.
 */
MlViewXMLDocument*
mlview_xml_document_clone (MlViewXMLDocument *a_original)
{	 
	MlViewXMLDocument * result ;
	MlViewFileDescriptor * file_desc ;
	gchar * file_path ;

	if ( a_original == NULL )
		return NULL ;
	file_desc = mlview_xml_document_get_file_descriptor (a_original) ;
	file_path = mlview_file_descriptor_get_file_path (file_desc);
	result = 
		mlview_xml_document_open 
		(file_path, PRIVATE (a_original)->app_context) ;
	return result ;
}

/**
 *Sets the file path associated to the current instance
 *of MlViewXMLDocument.
 *@param a_xml_doc the "this" pointer of the current object.
 *@param a_file_path the new file path.
 *
 */
void
mlview_xml_document_set_file_path (MlViewXMLDocument *a_xml_doc,
				   gchar *a_file_path)
{
	g_return_if_fail (a_xml_doc != NULL) ;
	g_return_if_fail (PRIVATE (a_xml_doc) != NULL) ;

	if (PRIVATE (a_xml_doc)->file_desc == NULL) 
        {
		PRIVATE (a_xml_doc)->file_desc =
			mlview_file_descriptor_new (a_file_path) ;
	} 
        else 
        {
		mlview_file_descriptor_set_file_path 
                        (PRIVATE (a_xml_doc)->file_desc,
                         a_file_path) ;
	}

	gtk_signal_emit (GTK_OBJECT (a_xml_doc),
			 gv_mlview_xml_document_signals [FILE_PATH_CHANGED]) ;
}


/**
 *Gets the file path associated to the current instance of
 *#MlViewXMLDocument.
 *@param a_xml_doc the "this" pointer of the current
 *instance.
 *@return the file path. Note that the caller should not free
 *the returned string!
 */
gchar *
mlview_xml_document_get_file_path (MlViewXMLDocument *a_xml_doc)
{
	g_return_val_if_fail (a_xml_doc 
                              && PRIVATE(a_xml_doc), NULL) ;

        if (PRIVATE (a_xml_doc)->file_desc == NULL)
        {
                return NULL ;
        }
        else
        {
                return mlview_file_descriptor_get_file_path 
                        (PRIVATE (a_xml_doc)->file_desc) ;
        }
}


/**
 *Adds a child node to the a_parent_node node. 
 *After adding the child node, this methods emits the following
 *signals: "child-node-added" and "document-changed".
 *
 *@param a_xml_node the node to add.
 *@param a_xml_doc the current instance of MlViewXMLDocument.
 *@param a_parent_node the parent of the node to add.
 *@return the newly added node or NULL if adding failed.
 */
xmlNode *
mlview_xml_document_add_child_node (MlViewXMLDocument *a_xml_doc,
				    xmlNode *a_parent_xml_node,
				    xmlNode *a_xml_node, 
				    gboolean a_subtree_required,
				    gboolean a_emit_signal)
{
	xmlNode * result = NULL  ;
	gchar *  validation_is_on = NULL ;

	g_return_val_if_fail (a_xml_doc != NULL, NULL) ;
	g_return_val_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_doc), NULL) ;
	g_return_val_if_fail (PRIVATE (a_xml_doc) != NULL, NULL) ;
	g_return_val_if_fail (a_parent_xml_node != NULL, NULL) ;
	g_return_val_if_fail ((a_parent_xml_node->type == XML_ELEMENT_NODE) 
			      || (a_parent_xml_node->type == XML_DOCUMENT_NODE), 
			      NULL) ;
	g_return_val_if_fail (a_xml_node != NULL, NULL) ;

	result = xmlAddChild (a_parent_xml_node, a_xml_node) ;
	g_return_val_if_fail (result != NULL, NULL) ;

	if (a_subtree_required == TRUE
	    && PRIVATE (a_xml_doc)->app_context
	    && mlview_app_context_settings_exist 
	    (PRIVATE (a_xml_doc)->app_context)) 
        {

		validation_is_on = 
			mlview_app_context_get_settings_value 
			(PRIVATE (a_xml_doc)->app_context,
			 MLVIEW_STG_K_IS_VALIDATION_ON) ;

		if (validation_is_on 
		    && !strcmp (validation_is_on, MLVIEW_STG_V_YES)
		    && result->type == XML_ELEMENT_NODE) 
                {

			enum MLVIEW_PARSING_UTILS_STATUS  status ;

			status =
				mlview_parsing_utils_build_required_attributes_list 
				(PRIVATE (a_xml_doc)->app_context, result) ;

			status =
				mlview_parsing_utils_build_required_children_tree 
				(PRIVATE (a_xml_doc)->app_context, &result) ;
		}

	}

	/*emit the proper signals*/
	if (a_emit_signal == TRUE) 
        {

		gtk_signal_emit 
			(GTK_OBJECT (a_xml_doc),
			 gv_mlview_xml_document_signals [CHILD_NODE_ADDED],
			 a_parent_xml_node, a_xml_node) ;

		gtk_signal_emit 
			(GTK_OBJECT (a_xml_doc),
			 gv_mlview_xml_document_signals [DOCUMENT_CHANGED]) ;

	}

	return result ;
}


/**
 *Cuts the node a_xml_node. The node is stored in the node clipboard
 *and then unlinked from the xml tree. 
 *After a successfull cut, this methods emits the following signals:
 *"node-cut" and "document-changed".
 *
 *@param a_xml_doc the current xml node
 *@return the xmlNode cut or NULL if something bad happened.
 */
xmlNode *
mlview_xml_document_cut_node (MlViewXMLDocument * a_xml_doc, 
			      xmlNode * a_xml_node,
			      gboolean a_emit_signal)
{
	xmlNode * parent_node = NULL , *result = NULL;

	g_return_val_if_fail (a_xml_doc != NULL, NULL) ;
	g_return_val_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_doc), NULL) ;
	g_return_val_if_fail (PRIVATE (a_xml_doc) != NULL, NULL) ;
	g_return_val_if_fail (PRIVATE (a_xml_doc)->xml_doc != NULL, NULL) ;
	g_return_val_if_fail (a_xml_node != NULL, NULL) ;

	g_return_val_if_fail (a_xml_node->doc 
			      == PRIVATE (a_xml_doc)->xml_doc,
			      NULL) ;

	parent_node = a_xml_node->parent ;

	g_return_val_if_fail (parent_node != NULL, NULL) ;

	mlview_xml_document_copy_node_to_clipboard 
		(a_xml_node, 
		 PRIVATE (a_xml_doc)->xml_doc) ;

	xmlUnlinkNode (a_xml_node) ;
	result = a_xml_node ;

	if (a_emit_signal == TRUE) {

		gtk_signal_emit (GTK_OBJECT (a_xml_doc),
				 gv_mlview_xml_document_signals [NODE_CUT],
				 parent_node, a_xml_node) ;

		gtk_signal_emit 
			(GTK_OBJECT (a_xml_doc),
			 gv_mlview_xml_document_signals [DOCUMENT_CHANGED]) ;
	}

	return result ;
}


/**
 *Sets a new content a_content to a_node.
 *After setting the content, this method emits the following
 *signals: "content-changed", "node-changed" and "document-changed".
 *
 *@param a_content the new content to set to a_node.
 *@param a_xml_document the current xml document.
 *@param a_node the node to set the new content to.
 *
 *@return the a_node with the new content or NULL if something bad happened. 
 */
xmlNode *
mlview_xml_document_set_node_content (MlViewXMLDocument * a_xml_document,
				      xmlNode *a_node, guchar * a_content,
				      enum MlViewEncoding a_enc,
				      gboolean a_emit_signal)
{
	guchar *utf8buf = NULL ;
	xmlNode * result = NULL ;

	g_return_val_if_fail (a_xml_document != NULL, NULL) ;
	g_return_val_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_document),
			      NULL) ;
	g_return_val_if_fail (a_node != NULL, NULL) ;

	if (a_enc == ISO8859_1) {
		enum MlViewStatus status = MLVIEW_OK ;

		status = 
			mlview_utils_isolat1_str_to_utf8 (a_content,
							  &utf8buf) ;

		if (status != MLVIEW_OK) {
			mlview_utils_trace_debug 
				("utf8 conversion failed") ;
			result = NULL ;
			goto release_resources ;
		}

	} else {
		utf8buf = NULL ;
	}

	xmlNodeSetContent (a_node, utf8buf) ;

	if (a_emit_signal) {

		gtk_signal_emit 
			(GTK_OBJECT (a_xml_document), 
			 gv_mlview_xml_document_signals[CONTENT_CHANGED], 
			 a_node) ;

		gtk_signal_emit (GTK_OBJECT (a_xml_document), 
				 gv_mlview_xml_document_signals[NODE_CHANGED], 
				 a_node) ;

		gtk_signal_emit 
			(GTK_OBJECT (a_xml_document), 
			 gv_mlview_xml_document_signals[DOCUMENT_CHANGED]) ;
	}

	result = a_node ;

 release_resources:

	if (utf8buf && utf8buf != a_content) {
		g_free (utf8buf) ;
		utf8buf = NULL ;
	}

	return result ;
}


/**
 *Sets the name of a_node to a_name. 
 *
 *@param a_xml_document the current xml document.
 *@param a_name the new name to set.
 *@param a_node the node to set the name on.
 *
 *@return a_node with it new name or NULL if something bad happened. 
 */
xmlNode *
mlview_xml_document_set_node_name (MlViewXMLDocument * a_xml_document,
				   xmlNode *a_node, guchar * a_name,
				   enum MlViewEncoding a_enc,
				   gboolean a_emit_signal)
{
	guchar *utf8_name = NULL ;
	enum MlViewStatus status = MLVIEW_OK ;

	g_return_val_if_fail (a_xml_document != NULL, 
			      NULL) ;
	g_return_val_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_document), 
			      NULL) ;
	g_return_val_if_fail (a_node != NULL, 
			      NULL) ;

	if (a_enc == ISO8859_1) {
		status =
			mlview_utils_isolat1_str_to_utf8 (a_name,
							  &utf8_name) ;

	} else if (a_enc == UTF8) {
		utf8_name = a_name ;
	}

	xmlNodeSetName (a_node, utf8_name) ;

	if (a_emit_signal) {

		gtk_signal_emit (GTK_OBJECT (a_xml_document),
				 gv_mlview_xml_document_signals [NAME_CHANGED],
				 a_node) ;
 
		gtk_signal_emit (GTK_OBJECT (a_xml_document), 
				 gv_mlview_xml_document_signals [NODE_CHANGED], 
				 a_node) ;

		gtk_signal_emit 
			(GTK_OBJECT (a_xml_document), 
			 gv_mlview_xml_document_signals [DOCUMENT_CHANGED]) ;	
	}

	if (utf8_name && utf8_name != a_name) {
		g_free (utf8_name) ;
		utf8_name = NULL ;
	}

	return a_node ;
}


enum MlViewStatus
mlview_xml_document_node_get_name (xmlNode *a_node,
				   enum MlViewEncoding a_enc,
				   guchar **a_outbuf)
{
	enum MlViewStatus status = MLVIEW_OK ;

	g_return_val_if_fail (a_node != NULL
			      && (a_node->type == XML_ELEMENT_NODE
				  || a_node->type == XML_DOCUMENT_NODE
				  || a_node->type == XML_PI_NODE)
			      && a_outbuf != NULL,
			      MLVIEW_BAD_PARAM_ERROR) ;

	if (a_node->name == NULL) {
		*a_outbuf = 0 ;
		return MLVIEW_OK ;
	}

	if (a_enc == ISO8859_1) {
		status =
			mlview_utils_utf8_str_to_isolat1 ((guchar*)a_node->name,
							  a_outbuf) ;
		
		if (status != MLVIEW_OK)
			return status ;

		status = MLVIEW_OK ;
		
	} else if (a_enc == UTF8) {

		*a_outbuf =
			g_strdup (a_node->name) ;
		status = MLVIEW_OK ;

	} else {
		status = MLVIEW_UNKNOWN_ENCODING_ERROR ;
	}

	return status ;
}


enum MlViewStatus
mlview_xml_document_node_get_fqn_len_as_isolat1 (xmlNode *a_node,
						 gint *a_len) {
	enum MlViewStatus status = MLVIEW_OK ;
	gchar *fqn = NULL ;
	gint len = 0 ;

	g_return_val_if_fail (a_node && a_len,
			      MLVIEW_BAD_PARAM_ERROR) ;

	if (a_node->ns
	    && a_node->ns->prefix
	    && mlview_utils_is_white_string
	    (a_node->ns->prefix) == FALSE) {

		fqn = g_strconcat (a_node->ns->prefix,
				   ":", a_node->name, NULL) ;
	} else {
		fqn = g_strdup (a_node->name) ;
	}

	if (fqn == NULL || *fqn == 0) {
		len = 0 ;
		status = MLVIEW_OK ;

	} else {
		status =
			mlview_utils_utf8_str_len_as_isolat1 (fqn, &len) ;
	}

	if (status == MLVIEW_OK)
		*a_len = len ;

	if (fqn) {
		g_free (fqn) ;
		fqn = NULL ;
	}

	return status ;
}


enum MlViewStatus
mlview_xml_document_node_get_fqn (xmlNode *a_node,
				  enum MlViewEncoding a_enc,
				  guchar **a_outbuf)
{
	enum MlViewStatus status = MLVIEW_OK ;
	gchar *fqn = NULL ;

	g_return_val_if_fail (a_node != NULL
			      && a_node->type == XML_ELEMENT_NODE
			      && a_outbuf != NULL,
			      MLVIEW_BAD_PARAM_ERROR) ;

	if (a_node->name == NULL) {
		*a_outbuf = 0 ;
		return MLVIEW_OK ;
	}

	if (a_node->ns 
	    && a_node->ns->prefix
	    && mlview_utils_is_white_string 
	    (a_node->ns->prefix) == FALSE) {

		fqn = g_strconcat (a_node->ns->prefix,
				   ":", a_node->name, NULL) ;
	} else {
		fqn = g_strdup (a_node->name) ;
	}

	if (a_enc == ISO8859_1) {
		status =
			mlview_utils_utf8_str_to_isolat1 (fqn, 
							  a_outbuf) ;

		if (status != MLVIEW_OK) {
			goto release_resources ;
		}


	} else if (a_enc == UTF8) {
		*a_outbuf = g_strdup (fqn) ;
	} else {
		status = MLVIEW_UNKNOWN_ENCODING_ERROR ;
		goto release_resources ;
	}


 release_resources:
	if (fqn) {
		g_free (fqn) ;
		fqn = NULL ;
	}

	return status ;
}

enum MlViewStatus
mlview_xml_document_node_get_content (xmlNode *a_node, 
				      enum MlViewEncoding a_enc,
				      guchar **a_outbuf)
{
	gchar *utf8_content = NULL ;
	enum MlViewStatus status = MLVIEW_OK ;

	g_return_val_if_fail (a_node != NULL 
			      && a_outbuf != NULL,
			      MLVIEW_BAD_PARAM_ERROR) ;
	
	utf8_content =
		xmlNodeGetContent (a_node) ;

	if (utf8_content == NULL) {
		*a_outbuf = 0 ;
		return MLVIEW_OK ;
	}

	if (a_enc == ISO8859_1) {
		status =
			mlview_utils_utf8_str_to_isolat1 (utf8_content,
							  a_outbuf) ;
		if (status != MLVIEW_OK) {
			goto release_resources ;
		}

		status = MLVIEW_OK ;

	} else if (a_enc == UTF8) {
		*a_outbuf = g_strdup (utf8_content) ;
		status = MLVIEW_OK ;
	} else {
		status = MLVIEW_UNKNOWN_ENCODING_ERROR ;
		goto release_resources ;
	}

 release_resources:
	if (utf8_content) {
		g_free (utf8_content) ;
		utf8_content = NULL ;
	}

	return status ;
}


/**
 *Inserts a_sibling_node as the previous sibling node of a_xml_node. 
 *
 *@param a_xml_node the current reference node. Must be not NULL.
 *@param a_xml_document the current xml document. Must be not NULL.
 *@param a_sibling_node a the new sibling node to insert. Must be not NULL.
 *
 *@return the newly insterted sibling (a_sibling_node) or 
 *NULL if something BAD happened. 
 */
xmlNode *
mlview_xml_document_insert_prev_sibling_node (MlViewXMLDocument *a_xml_document,
					      xmlNode *a_sibling_node, 
					      xmlNode *a_xml_node,
					      gboolean a_subtree_required,
					      gboolean a_emit_signal)
{
	xmlNode * result = NULL ;
	gchar *  validation_is_on = NULL ;

	g_return_val_if_fail (a_xml_document != NULL, NULL) ;
	g_return_val_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_document), NULL) ;
	g_return_val_if_fail (PRIVATE (a_xml_document) != NULL, NULL) ;
	g_return_val_if_fail (a_sibling_node != NULL, NULL) ;
	g_return_val_if_fail (a_xml_node != NULL, NULL) ;

	result = xmlAddPrevSibling (a_sibling_node, a_xml_node) ;

	if (a_subtree_required == TRUE
	    && PRIVATE (a_xml_document)->app_context
	    && mlview_app_context_settings_exist 
	    (PRIVATE (a_xml_document)->app_context)) {

		validation_is_on =
			mlview_app_context_get_settings_value 
			(PRIVATE (a_xml_document)->app_context,
			 MLVIEW_STG_K_IS_VALIDATION_ON) ;

		if (validation_is_on 
		    && !strcmp (validation_is_on, MLVIEW_STG_V_YES)
		    && result->type == XML_ELEMENT_NODE) {

			enum MLVIEW_PARSING_UTILS_STATUS  status ;
			status = 
				mlview_parsing_utils_build_required_attributes_list 
				(PRIVATE (a_xml_document)->app_context,
				 result) ;

			status = 
				mlview_parsing_utils_build_required_children_tree 
				(PRIVATE (a_xml_document)->app_context,  
				 &result) ;
		}			
	}

	if (a_emit_signal == TRUE) {

		gtk_signal_emit 
			(GTK_OBJECT (a_xml_document),
			 gv_mlview_xml_document_signals[PREV_SIBLING_NODE_INSERTED],
			 a_sibling_node,
			 a_xml_node) ;

		gtk_signal_emit 
			(GTK_OBJECT (a_xml_document),
			 gv_mlview_xml_document_signals[DOCUMENT_CHANGED]) ;

	}

	return result ;
}


/**
 *Inserts a_sibling_node as  the next sibling of a_xml_node. 
 *
 *@param a_xml_node the reference xml node.
 *@param a_xml_document the current xml document.
 *@param a_sibling_node the new sibling node to insert.
 *
 *
 *@return the newly inserted next sibling node (a_sibling_node) or 
 */
xmlNode *
mlview_xml_document_insert_next_sibling_node (MlViewXMLDocument *a_xml_document,
					      xmlNode *a_sibling_node, 
					      xmlNode *a_xml_node,
					      gboolean a_subtree_required,
					      gboolean a_emit_signal)
{
	xmlNode * result = NULL ;
	gchar *  validation_is_on = NULL ;

	g_return_val_if_fail (a_xml_document != NULL, NULL) ;
	g_return_val_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_document), NULL) ;
	g_return_val_if_fail (PRIVATE (a_xml_document) != NULL, NULL) ;
	g_return_val_if_fail (a_sibling_node != NULL, NULL) ;
	g_return_val_if_fail (a_xml_node != NULL, NULL) ;
	
	result = xmlAddNextSibling (a_sibling_node, a_xml_node) ;
	
	if (a_subtree_required == TRUE
	    && PRIVATE (a_xml_document)->app_context
	    && mlview_app_context_settings_exist (PRIVATE (a_xml_document)->app_context)) {
		validation_is_on =
			mlview_app_context_get_settings_value (PRIVATE (a_xml_document)->app_context,
							       MLVIEW_STG_K_IS_VALIDATION_ON) ;

		if (validation_is_on 
		    && !strcmp (validation_is_on, MLVIEW_STG_V_YES)
		    && result->type == XML_ELEMENT_NODE) {

			enum MLVIEW_PARSING_UTILS_STATUS  status ;

			status =
				mlview_parsing_utils_build_required_attributes_list 
                                (PRIVATE (a_xml_document)->app_context, 
                                 result) ;
			status = 
				mlview_parsing_utils_build_required_children_tree 
				(PRIVATE (a_xml_document)->app_context, &result) ;
		}
			
	}

	if (a_emit_signal == TRUE) {

		gtk_signal_emit 
			(GTK_OBJECT (a_xml_document),
			 gv_mlview_xml_document_signals [NEXT_SIBLING_NODE_INSERTED],
			 a_sibling_node,
			 a_xml_node) ;

		gtk_signal_emit 
			(GTK_OBJECT (a_xml_document),
			 gv_mlview_xml_document_signals [DOCUMENT_CHANGED]) ;
	}

	return result ;
}


void
mlview_xml_document_paste_node_as_child (MlViewXMLDocument *a_xml_document,
					 xmlNode *a_parent_node,
					 gboolean a_emit_signal)
{
	xmlNode * xml_node = NULL ;

	g_return_if_fail (a_xml_document != NULL) ;
	g_return_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_document)) ;
	g_return_if_fail (PRIVATE (a_xml_document)) ;
	g_return_if_fail (a_parent_node != NULL) ;

	xml_node = 
		mlview_xml_document_get_node_from_clipboard 
		(PRIVATE (a_xml_document)->xml_doc) ;

	g_return_if_fail (xml_node != NULL) ;

	mlview_xml_document_add_child_node (a_xml_document, a_parent_node, 
					    xml_node, FALSE, a_emit_signal) ;
}


void
mlview_xml_document_paste_node_as_sibling (MlViewXMLDocument * a_xml_document, 
					   xmlNode *a_parent_node,
					   xmlNode * a_sibling_node, 
					   gboolean a_previous,
					   gboolean a_emit_signal)
{
	xmlNode *xml_node = NULL ;

	g_return_if_fail (a_xml_document != NULL) ;
	g_return_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_document)) ;
	g_return_if_fail (PRIVATE (a_xml_document) != NULL) ;
	g_return_if_fail (a_parent_node != NULL) ;
	g_return_if_fail (a_sibling_node != NULL) ;

	xml_node = 
		mlview_xml_document_get_node_from_clipboard 
		(PRIVATE (a_xml_document)->xml_doc) ;

	g_return_if_fail (xml_node != NULL) ;

	if (a_previous == FALSE) {

		mlview_xml_document_insert_next_sibling_node (a_xml_document,
							      a_sibling_node, 
							      xml_node, FALSE,
							      a_emit_signal) ;
	} else {
		mlview_xml_document_insert_prev_sibling_node (a_xml_document,
							      a_sibling_node, 
							      xml_node, FALSE,
							      a_emit_signal) ;
	}
}



gboolean
mlview_xml_document_needs_saving (MlViewXMLDocument *a_doc)
{
	MlViewFileDescriptor *file_desc ;
	gboolean is_modified ;

	g_return_val_if_fail (a_doc != NULL, FALSE) ;
	g_return_val_if_fail (PRIVATE(a_doc) != NULL, FALSE) ;
	
	file_desc = mlview_xml_document_get_file_descriptor (a_doc) ;
	if ( file_desc == NULL )
		return TRUE ;
	
	if ( mlview_file_descriptor_is_modified (file_desc, &is_modified) )
		return FALSE ;
	return is_modified ;
}


void 
mlview_xml_document_destroy (GtkObject *a_xml_doc)
{
	MlViewXMLDocument * xml_doc = NULL ;

	g_return_if_fail (a_xml_doc != NULL) ;
	g_return_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_doc)) ;
	
	xml_doc = MLVIEW_XML_DOCUMENT (a_xml_doc) ;
        
	if ( PRIVATE (xml_doc) == NULL )
		return ;

	if ( PRIVATE(xml_doc)->file_desc != NULL ) 
	{
		mlview_file_descriptor_destroy (PRIVATE (xml_doc)->file_desc) ;
		PRIVATE (xml_doc)->file_desc = NULL ;
        }

        /*
         *Free the nodes contained in the clipboard
         *before freeing the document that contain
         *them or else, the nodes will have a dangling
         *pointer to a freed document. That could lead
         *to invalid read or invalid write and thus to
         *stability problems.
         */
        mlview_xml_document_unref_clipboard () ;

	if ( PRIVATE (xml_doc)->xml_doc != NULL )
	{
		xmlFreeDoc (PRIVATE (xml_doc)->xml_doc) ;

		PRIVATE (xml_doc)->xml_doc = NULL ;
	}

        

	g_free (PRIVATE (xml_doc)) ;
	PRIVATE (xml_doc) = NULL ;

        gtk_object_destroy (a_xml_doc) ;
}


/**
 *Increments the reference count associated to this instance.
 *To decrement that reference count, please use
 *mlview_xml_document_unref().
 *
 *@param a_this the "this pointer".
 */
void
mlview_xml_document_ref (MlViewXMLDocument *a_this)
{
        g_return_if_fail (a_this
                          && MLVIEW_IS_XML_DOCUMENT (a_this)
                          && PRIVATE (a_this)) ;

        gtk_object_ref (GTK_OBJECT (a_this)) ;
}


/**
 *Decrements the reference count associated to the current
 *instance of #MlViewXMLDocument. If the reference count reaches
 *zero, the current instance of #MlViewXMLDocument is destroyed.
 *
 *@param a_this in out parameter. Points to the this pointer.
 *If the ref count reached zero and the current instance has been
 *destroyed, *a_this is set to NULL.
 */
void
mlview_xml_document_unref (MlViewXMLDocument *a_this)
{
        g_return_if_fail (a_this 
                          && MLVIEW_IS_XML_DOCUMENT (a_this)
                          && PRIVATE (a_this)) ;

        gtk_object_unref (GTK_OBJECT (a_this)) ;
}

                                                                                
/**
 *Gets the file descriptor asscociated to the current instance
 *of #MlViewXMLDocument.
 *
 *@param a_xml_doc the "this pointer".
 *@return the file descriptor associated to the document.
 */
MlViewFileDescriptor *
mlview_xml_document_get_file_descriptor (MlViewXMLDocument *a_xml_doc)
{
	g_return_val_if_fail (a_xml_doc != NULL, NULL) ;
	g_return_val_if_fail (PRIVATE (a_xml_doc) != NULL, NULL) ;
	
	return PRIVATE (a_xml_doc)->file_desc ;
}

/**
 *Sets the file descriptor associated to the document.
 *
 *@param a_xml_doc the this pointer.
 *@param a_file_desc the new file descriptor to set.
 */
void 
mlview_xml_document_set_file_descriptor (MlViewXMLDocument *a_xml_doc, 
                                         MlViewFileDescriptor *a_file_desc)
{
	g_return_if_fail (a_xml_doc != NULL) ;
	g_return_if_fail (PRIVATE(a_xml_doc) != NULL) ;
	
	PRIVATE (a_xml_doc)->file_desc = a_file_desc ;
}

/**
 *Gets the native underlaying libxml2 document.
 *
 *@param a_xml_doc the "this pointer".
 *@return the underlying native libxml2 document.
 */
xmlDocPtr
mlview_xml_document_get_xml_document (MlViewXMLDocument *a_xml_doc)
{
	g_return_val_if_fail (a_xml_doc != NULL, NULL) ;
	g_return_val_if_fail (PRIVATE(a_xml_doc) != NULL, NULL) ;
	
	return PRIVATE (a_xml_doc)->xml_doc ;
}


/**
 *Sets a new underlying native libxml2 document.
 *
 *@param a_doc the "this pointer".
 *@param a_mem_doc a pointer to the new xmlDoc to set.
 */
void
mlview_xml_document_set_xml_document (MlViewXMLDocument *a_doc, 
				      xmlDocPtr a_mem_doc)
{
	g_return_if_fail (a_doc != NULL) ;
	g_return_if_fail (PRIVATE(a_doc) != NULL) ;

	PRIVATE (a_doc)->xml_doc = a_mem_doc ;	
}
